﻿/////////////////////////////////////////////////////////////////////////////////////////
//                              INTEL CONFIDENTIAL
//              Copyright 2018 Intel Corporation All Rights Reserved.
/////////////////////////////////////////////////////////////////////////////////////////
//
// The source code contained or described herein and all documents related to the source 
// code ("Material") are owned by Intel Corporation or its suppliers or licensors. Title 
// to the Material remains with Intel Corporation or its suppliers and licensors. The 
// Material may contain trade secrets and proprietary and confidential information of 
// Intel Corporation and its suppliers and licensors, and is protected by worldwide 
// copyright and trade secret laws and treaty provisions. No part of the Material may be 
// used, copied, reproduced, modified, published, uploaded, posted, transmitted, distributed, 
// or disclosed in any way without Intel’s prior express written permission. 
//
// No license under any patent, copyright, trade secret or other intellectual property 
// right is granted to or conferred upon you by disclosure or delivery of the Materials, 
// either expressly, by implication, inducement, estoppel or otherwise. Any license under
// such intellectual property rights must be express and approved by Intel in writing.
//
//////////////////////////////////////////////////////////////////////////

/************************************************************************************************************************
    Example for using the KS control utils to get and set exposure priority:

    //Create the CameraControl for exposure priority (KSPROPERTY_CAMERACONTROL_AUTO_EXPOSURE_PRIORITY = 19)
    CameraControl exposurePriority = new CameraControl(19, _mediaCapture.VideoDeviceController);

    // Get the current exposure priority value
    var expPriorityValue = exposurePriority.Get();

    // Set the new exposure priority value.
    var newExposurePriorityValue = false;
    exposurePriority.Set(newExposurePriorityValue);

    Be aware that your application MUST be in ExclusiveControl mode (SharingMode = MediaCaptureSharingMode.ExclusiveControl,). 
    If it is in shared the control will not reach the device but you will not get an exception.

************************************************************************************************************************/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace KsControlUtils
{
    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct KSIDENTIFIER
    {
        public Guid set;
        public UInt32 id;
        public UInt32 flags;
    };

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct PROPSETID_VIDCAP
    {
        public KSIDENTIFIER ksproperty;
        public Int32 value;
        public UInt32 flags;
        public UInt32 capabilities;
        public UInt32 padding;
    };

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct KSPROPERTY_MEMBERSHEADER
    {
        public UInt32 MembersFlags;
        public UInt32 MembersSize;
        public UInt32 MembersCount;
        public UInt32 Flags;
    };

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct KSPROPERTY_BOUNDS_LONG
    {
        public UInt32 Minimum;
        public UInt32 Maximum;
    };

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public struct KSPROPERTY_STEPPING_LONG
    {
        public UInt32 SteppingDelta;
        public UInt32 Reserved;
        public KSPROPERTY_BOUNDS_LONG Bounds;
    };

    [StructLayout(LayoutKind.Sequential)]
    public struct KSPROPERTY_DESCRIPTION
    {
        public UInt32       AccessFlags;
        public UInt32       DescriptionSize;
        public KSIDENTIFIER PropTypeSet;
        public UInt32       MembersListCount;
        public UInt32       Reserved;
    };

    [StructLayout(LayoutKind.Sequential)]
    public struct VIDCAP_BASIC_SUPPORT
    {
        public KSPROPERTY_DESCRIPTION desc;
        public KSPROPERTY_MEMBERSHEADER hdr;
        public Int32 step;
        public Int32 min;
        public Int32 max;
        public Int32 reserved;
    };

    [StructLayout(LayoutKind.Sequential)]
    public struct VIDCAP_DEFAULT_VALUES
    {
        public KSPROPERTY_DESCRIPTION desc;
        public KSPROPERTY_MEMBERSHEADER hdr;
        public Int32 def;
        public Int32 reserved;
    };

    public struct KSPROPERTY_CAMERACONTROL_S
    {
        public KSIDENTIFIER Property;
        public Int32 Value;                   // value to get or set
        public UInt32 Flags;                   // KSPROPERTY_CAMERACONTROL_FLAGS_*
        public UInt32 Capabilities;            // KSPROPERTY_CAMERACONTROL_FLAGS_*
        public UInt32 padding;
    };
	
	public class DeviceProperty
    {
        protected string str;
        protected uint id;
        protected Windows.Media.Devices.VideoDeviceController controller;

        protected DeviceProperty(string theSet, uint theId, Windows.Media.Devices.VideoDeviceController deviceController)
        {
            id = theId;
            str = "{" + theSet + "} " + id;
            controller = deviceController;
        }

        protected byte[] PdoToBuffer<T>(T obj)
        {
            int len = Marshal.SizeOf<T>();
            byte[] arr = new byte[len];
            IntPtr ptr = Marshal.AllocHGlobal(len);
            Marshal.StructureToPtr<T>(obj, ptr, true);
            Marshal.Copy(ptr, arr, 0, len);
            Marshal.FreeHGlobal(ptr);
            return arr;
        }

        public static T ByteArrayToStructure<T>(byte[] arr)
        {
            T str = default(T);
            int size = Marshal.SizeOf<T>();
            IntPtr ptr = Marshal.AllocHGlobal(size);
            Marshal.Copy(arr, 0, ptr, size);
            str = Marshal.PtrToStructure<T>(ptr);
            Marshal.FreeHGlobal(ptr);
            return str;
        }
    }

    public class PropsetidVidcapControl : DeviceProperty
    {
        public PropsetidVidcapControl(string theSet, uint theId, Windows.Media.Devices.VideoDeviceController deviceController)
            :base(theSet, theId, deviceController)
        {
            set = new Guid(theSet);
        }

        public bool Set(int value)
        {
            try
            {
                PROPSETID_VIDCAP vidCap = new PROPSETID_VIDCAP();
                vidCap.ksproperty.set = set;
                vidCap.ksproperty.id = id;
                vidCap.ksproperty.flags = 2;
                vidCap.flags = 0x2;
                vidCap.capabilities = 0;
                vidCap.value = value;

                byte[] data = PdoToBuffer<PROPSETID_VIDCAP>(vidCap);
                //byte[] ksId = PdoToBuffer<KSIDENTIFIER>(vidCap.ksproperty);
                controller.SetDevicePropertyByExtendedId(data, data);
                //controller.SetDeviceProperty(str, data);
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Failed to set with exception " + ex.ToString());
                return false;
            }
            return true;
        }

        public int Get()
        {
            try
            {
                PROPSETID_VIDCAP vidCap = new PROPSETID_VIDCAP();
                vidCap.ksproperty.set = set;
                vidCap.ksproperty.id = id;
                vidCap.ksproperty.flags = 1;
                vidCap.flags = 0;
                vidCap.capabilities = 0;
                vidCap.value = 0;

                byte[] data = PdoToBuffer<PROPSETID_VIDCAP>(vidCap);
                int size = Marshal.SizeOf<KSPROPERTY_CAMERACONTROL_S>();
                var result = controller.GetDevicePropertyByExtendedId(data, /*size of your expected returned struct in bytes*/(uint)size);

                Debug.WriteLine(result.Status);
                byte[] propertyData = result.Value as byte[];
                Debug.Assert(propertyData != null && propertyData.Length > 0);
                KSPROPERTY_CAMERACONTROL_S resultValue = ByteArrayToStructure<KSPROPERTY_CAMERACONTROL_S>(propertyData);
                return resultValue.Value;
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Failed to set with exception " + ex.ToString());
                return 0;
            }
        }

        public VIDCAP_BASIC_SUPPORT GetBasicSupport()
        {
            try
            {
                PROPSETID_VIDCAP vidCap = new PROPSETID_VIDCAP();
                vidCap.ksproperty.set = set;
                vidCap.ksproperty.id = id;
                vidCap.ksproperty.flags = 0x200;
                vidCap.flags = 0;
                vidCap.capabilities = 0;
                vidCap.value = 0;

                byte[] data = PdoToBuffer<PROPSETID_VIDCAP>(vidCap);
                int size = Marshal.SizeOf<VIDCAP_BASIC_SUPPORT>();
                var result = controller.GetDevicePropertyByExtendedId(data, /*size of your expected returned struct in bytes*/(uint)size);

                Debug.WriteLine(result.Status);
                byte[] propertyData = result.Value as byte[];
                Debug.Assert(propertyData != null && propertyData.Length > 0);
                VIDCAP_BASIC_SUPPORT resultValue = ByteArrayToStructure<VIDCAP_BASIC_SUPPORT>(propertyData);
                return resultValue;
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Failed to set with exception " + ex.ToString());
                return new VIDCAP_BASIC_SUPPORT();
            }
        }

        public VIDCAP_DEFAULT_VALUES GetDefault()
        {
            try
            {
                PROPSETID_VIDCAP vidCap = new PROPSETID_VIDCAP();
                vidCap.ksproperty.set = set;
                vidCap.ksproperty.id = id;
                vidCap.ksproperty.flags = 0x10000;
                vidCap.flags = 0;
                vidCap.capabilities = 0;
                vidCap.value = 0;

                byte[] data = PdoToBuffer<PROPSETID_VIDCAP>(vidCap);
                int size = Marshal.SizeOf<VIDCAP_DEFAULT_VALUES>();
                var result = controller.GetDevicePropertyByExtendedId(data, /*size of your expected returned struct in bytes*/(uint)size);

                Debug.WriteLine(result.Status);
                byte[] propertyData = result.Value as byte[];
                Debug.Assert(propertyData != null && propertyData.Length > 0);
                VIDCAP_DEFAULT_VALUES resultValue = ByteArrayToStructure<VIDCAP_DEFAULT_VALUES>(propertyData);
                return resultValue;
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Failed to set with exception " + ex.ToString());
                return new VIDCAP_DEFAULT_VALUES();
            }
        }

        Guid set;
    }
    public class VideoProcAmp : PropsetidVidcapControl
    {
        public VideoProcAmp(uint id, Windows.Media.Devices.VideoDeviceController deviceController)
            :base("C6E13360-30AC-11d0-A18C-00A0C9118956", id, deviceController)
        {
        }
    }

    public class CameraControl : PropsetidVidcapControl
    {
        public CameraControl(uint id, Windows.Media.Devices.VideoDeviceController deviceController)
            : base("C6E13370-30AC-11d0-A18C-00A0C9118956", id, deviceController)
        {
        }
    }
}